import unittest

from leeger.util.GeneralUtil import GeneralUtil


class TestGeneralUtil(unittest.TestCase):
    def test_filter_None(self):
        l = [None, "test", None]
        response = GeneralUtil.filter(value=None, list_=l)
        self.assertEqual(["test"], response)

    def test_filter_emptyList(self):
        l = list()
        response = GeneralUtil.filter(value=None, list_=l)
        self.assertEqual([], response)

    def test_warnForUnusedKwargs_happyPath(self):
        kwargs = {"a": 1, "b": 2}
        with self.assertLogs() as captured:
            GeneralUtil.warnForUnusedKwargs(kwargs)
        self.assertEqual(2, len(captured.records))
        self.assertEqual("Keyword argument 'a' unused.", captured.records[0].getMessage())
        self.assertEqual("Keyword argument 'b' unused.", captured.records[1].getMessage())

    def test_safeSum_happyPath(self):
        response = GeneralUtil.safeSum(None, 1, 1)
        self.assertEqual(2, response)

        response = GeneralUtil.safeSum(1, None, 1)
        self.assertEqual(2, response)

        response = GeneralUtil.safeSum(1, 1, None)
        self.assertEqual(2, response)

        response = GeneralUtil.safeSum(None, None, None)
        self.assertIsNone(response)

        response = GeneralUtil.safeSum(1, 1, 1)
        self.assertEqual(3, response)



from typing import Optional

from leeger.calculator.parent.AllTimeCalculator import AllTimeCalculator
from leeger.calculator.year_calculator.GameOutcomeYearCalculator import GameOutcomeYearCalculator
from leeger.decorator.validators import validateLeague
from leeger.model.filter import AllTimeFilters
from leeger.model.league.League import League
from leeger.util.Deci import Deci
from leeger.util.navigator.LeagueNavigator import LeagueNavigator


class GameOutcomeAllTimeCalculator(AllTimeCalculator):

    @classmethod
    @validateLeague
    def getWins(cls, league: League, **kwargs) -> dict[str, Optional[int]]:
        """
        Returns the number of wins for each team in the given League.
        Returns None for an Owner if they have no games played in the range.

        Example response:
            {
            "someOwnerId": 18,
            "someOtherOwnerId": 21,
            "yetAnotherOwnerId": 17,
            ...
            }
        """
        return cls._addAndCombineResults(league, GameOutcomeYearCalculator.getWins, **kwargs)

    @classmethod
    @validateLeague
    def getLosses(cls, league: League, **kwargs) -> dict[str, Optional[int]]:
        """
        Returns the number of losses for each team in the given League.
        Returns None for an Owner if they have no games played in the range.

        Example response:
            {
            "someOwnerId": 18,
            "someOtherOwnerId": 21,
            "yetAnotherOwnerId": 17,
            ...
            }
        """
        return cls._addAndCombineResults(league, GameOutcomeYearCalculator.getLosses, **kwargs)

    @classmethod
    @validateLeague
    def getTies(cls, league: League, **kwargs) -> dict[str, Optional[int]]:
        """
        Returns the number of ties for each team in the given League.
        Returns None for an Owner if they have no games played in the range.

        Example response:
            {
            "someOwnerId": 1,
            "someOtherOwnerId": 0,
            "yetAnotherOwnerId": 2,
            ...
            }
        """
        return cls._addAndCombineResults(league, GameOutcomeYearCalculator.getTies, **kwargs)

    @classmethod
    @validateLeague
    def getWinPercentage(cls, league: League, **kwargs) -> dict[str, Optional[Deci]]:
        """
        Returns the win percentage for each team in the given League.
        Returns None for an Owner if they have no games played in the range.
        Win percentage will be represented as a non-rounded decimal.

        Example:
            33% win percentage -> Deci("0.3333333333333333333333333333")
            100% win percentage -> Deci("1.0")

        Example response:
            {
            "someOwnerId": Deci("0.85"),
            "someOtherOwnerId": Deci("0.1156"),
            "yetAnotherOwnerId": Deci("0.72"),
            ...
            }
        """
        ownerIdAndWinPercentage = dict()
        ownerIdAndWins = GameOutcomeAllTimeCalculator.getWins(league, **kwargs)
        ownerIdAndLosses = GameOutcomeAllTimeCalculator.getLosses(league, **kwargs)
        ownerIdAndTies = GameOutcomeAllTimeCalculator.getTies(league, **kwargs)
        ownerIdAndLeagueMedianWins = GameOutcomeAllTimeCalculator.getLeagueMedianWins(league, **kwargs)

        for ownerId in [owner.id for owner in league.owners]:
            numberOfWins = ownerIdAndWins[ownerId]
            numberOfLosses = ownerIdAndLosses[ownerId]
            numberOfTies = ownerIdAndTies[ownerId]
            numberOfLeagueMedianWins = ownerIdAndLeagueMedianWins[ownerId]

            if None in (numberOfWins, numberOfLosses, numberOfTies, numberOfLeagueMedianWins):
                ownerIdAndWinPercentage[ownerId] = None
            else:
                filters = AllTimeFilters.getForLeague(league, **kwargs)
                numberOfGamesPlayed = LeagueNavigator.getNumberOfGamesPlayed(league,
                                                                             filters,
                                                                             countMultiWeekMatchupsAsOneGame=True,
                                                                             countLeagueMedianGamesAsTwoGames=True)[
                    ownerId]
                totalWins = numberOfWins + numberOfLeagueMedianWins
                ownerIdAndWinPercentage[ownerId] = (Deci(totalWins) + (Deci("0.5") * Deci(numberOfTies))) / Deci(
                    numberOfGamesPlayed)

        return ownerIdAndWinPercentage

    @classmethod
    @validateLeague
    def getWAL(cls, league: League, **kwargs) -> dict[str, Optional[Deci]]:
        """
        WAL is "Wins Against the League"
        Formula: (Number of Wins * 1) + (Number of Ties * 0.5)

        Returns the number of Wins Against the League for each team in the given League.
        Returns None for an Owner if they have no games played in the range.
        If applicable, League Median Wins are counted towards this stat.

        Example response:
            {
            "someOwnerId": Deci("18.5"),
            "someOtherOwnerId": Deci("21.0"),
            "yetAnotherOwnerId": Deci("27.5"),
            ...
            }
        """

        ownerIdAndWAL = dict()
        ownerIdAndWins = GameOutcomeAllTimeCalculator.getWins(league, **kwargs)
        ownerIdAndTies = GameOutcomeAllTimeCalculator.getTies(league, **kwargs)
        ownerIdAndLeagueMedianWins = GameOutcomeAllTimeCalculator.getLeagueMedianWins(league, **kwargs)

        for ownerId in [owner.id for owner in league.owners]:
            wins = ownerIdAndWins[ownerId]
            ties = ownerIdAndTies[ownerId]
            leagueMedianWins = ownerIdAndLeagueMedianWins[ownerId]
            if None in (wins, ties):
                ownerIdAndWAL[ownerId] = None
            else:
                ownerIdAndWAL[ownerId] = Deci(wins) + (Deci("0.5") * Deci(ties))
                if ownerIdAndLeagueMedianWins[ownerId] is not None:
                    ownerIdAndWAL[ownerId] += Deci(leagueMedianWins)

        return ownerIdAndWAL

    @classmethod
    @validateLeague
    def getWALPerGame(cls, league: League, **kwargs) -> dict[str, Optional[Deci]]:
        """
        Returns the number of Wins Against the League per game for each owner in the given League.
        Returns None for an Owner if they have no games played in the range.

        Example response:
            {
            "someOwnerId": Deci("0.7"),
            "someOtherOwnerId": Deci("0.29"),
            "yetAnotherOwnerId": Deci("0.48"),
            ...
            }
        """
        ownerIdAndWAL = cls.getWAL(league, **kwargs)
        ownerIdAndNumberOfGamesPlayed = LeagueNavigator.getNumberOfGamesPlayed(league,
                                                                               AllTimeFilters.getForLeague(league,
                                                                                                           **kwargs),
                                                                               countMultiWeekMatchupsAsOneGame=True,
                                                                               countLeagueMedianGamesAsTwoGames=True)

        ownerIdAndWALPerGame = dict()
        allOwnerIds = LeagueNavigator.getAllOwnerIds(league)
        for ownerId in allOwnerIds:
            if ownerIdAndNumberOfGamesPlayed[ownerId] == 0:
                ownerIdAndWALPerGame[ownerId] = None
            else:
                ownerIdAndWALPerGame[ownerId] = ownerIdAndWAL[ownerId] / ownerIdAndNumberOfGamesPlayed[ownerId]

        return ownerIdAndWALPerGame

    @classmethod
    @validateLeague
    def getLeagueMedianWins(cls, league: League, **kwargs) -> dict[str, Optional[Deci]]:
        """
        Returns the number of league median wins for each Owner in the given League.
        Returns None for an Owner if they have no games played in the range.
        If there is a league median tie, then 0.5 wins is given to each team in the tie.
        This calculation is only run for regular season weeks.

        Example response:
            {
            "someTeamId": Deci("18.0"),
            "someOtherTeamId": Deci("21.0"),
            "yetAnotherTeamId": Deci("17.5"),
            ...
            }
        """
        return cls._addAndCombineResults(league, GameOutcomeYearCalculator.getLeagueMedianWins, **kwargs)

    @classmethod
    @validateLeague
    def getOpponentLeagueMedianWins(cls, league: League, **kwargs) -> dict[str, Optional[Deci]]:
        """
        Returns the number of league median wins for each Owner's opponents in the given League.
        Returns None for an Owner if they have no games played in the range.
        If there is a league median tie, then 0.5 wins is given to each team in the tie.
        This calculation is only run for regular season weeks.

        Example response:
            {
            "someTeamId": Deci("18.0"),
            "someOtherTeamId": Deci("21.0"),
            "yetAnotherTeamId": Deci("17.5"),
            ...
            }
        """
        return cls._addAndCombineResults(league, GameOutcomeYearCalculator.getOpponentLeagueMedianWins, **kwargs)



import unittest

from leeger.model.league import YearSettings
from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year
from leeger.model.stat.AllTimeStatSheet import AllTimeStatSheet
from leeger.model.stat.YearStatSheet import YearStatSheet
from test.helper.prototypes import getNDefaultOwnersAndTeams


class TestStatSheet(unittest.TestCase):

    def test_leagueStatSheet(self):
        from leeger.util.stat_sheet import leagueStatSheet
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1, teamBScore=2)
        week = Week(weekNumber=1, matchups=[matchup])
        year = Year(yearNumber=2000, teams=teams, weeks=[week])

        league = League(name="TEST", owners=owners, years=[year])

        leagueStatSheet = leagueStatSheet(league)

        self.assertIsInstance(leagueStatSheet, AllTimeStatSheet)
        self.assertIsInstance(leagueStatSheet.gamesPlayed, dict)
        self.assertIsInstance(leagueStatSheet.wins, dict)
        self.assertIsInstance(leagueStatSheet.losses, dict)
        self.assertIsInstance(leagueStatSheet.ties, dict)
        self.assertIsInstance(leagueStatSheet.winPercentage, dict)
        self.assertIsInstance(leagueStatSheet.wal, dict)
        self.assertIsInstance(leagueStatSheet.walPerGame, dict)

        self.assertIsInstance(leagueStatSheet.awal, dict)
        self.assertIsInstance(leagueStatSheet.awalPerGame, dict)
        self.assertIsInstance(leagueStatSheet.opponentAWAL, dict)
        self.assertIsInstance(leagueStatSheet.opponentAWALPerGame, dict)

        self.assertIsInstance(leagueStatSheet.smartWins, dict)
        self.assertIsInstance(leagueStatSheet.smartWinsPerGame, dict)
        self.assertIsInstance(leagueStatSheet.opponentSmartWins, dict)
        self.assertIsInstance(leagueStatSheet.opponentSmartWinsPerGame, dict)

        self.assertIsInstance(leagueStatSheet.pointsScored, dict)
        self.assertIsInstance(leagueStatSheet.pointsScoredPerGame, dict)
        self.assertIsInstance(leagueStatSheet.opponentPointsScored, dict)
        self.assertIsInstance(leagueStatSheet.opponentPointsScoredPerGame, dict)

        self.assertIsInstance(leagueStatSheet.scoringShare, dict)
        self.assertIsInstance(leagueStatSheet.opponentScoringShare, dict)
        self.assertIsInstance(leagueStatSheet.maxScoringShare, dict)
        self.assertIsInstance(leagueStatSheet.minScoringShare, dict)

        self.assertIsInstance(leagueStatSheet.maxScore, dict)
        self.assertIsInstance(leagueStatSheet.minScore, dict)

        self.assertIsInstance(leagueStatSheet.scoringStandardDeviation, dict)

        self.assertIsInstance(leagueStatSheet.plusMinus, dict)

        self.assertIsInstance(leagueStatSheet.adjustedTeamScore, dict)
        self.assertIsInstance(leagueStatSheet.adjustedTeamSuccess, dict)
        self.assertIsInstance(leagueStatSheet.adjustedTeamLuck, dict)

        # check optional stats are None
        self.assertIsNone(leagueStatSheet.leagueMedianWins)

    def test_leagueStatSheet_leagueMedianGamesIsTrueInAnyYearSettings(self):
        from leeger.util.stat_sheet import leagueStatSheet
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1, teamBScore=2)
        week = Week(weekNumber=1, matchups=[matchup])
        yearSettings = YearSettings(leagueMedianGames=True)
        year = Year(yearNumber=2000, teams=teams, weeks=[week], yearSettings=yearSettings)

        league = League(name="TEST", owners=owners, years=[year])

        leagueStatSheet = leagueStatSheet(league)

        self.assertIsInstance(leagueStatSheet, AllTimeStatSheet)
        self.assertIsInstance(leagueStatSheet.leagueMedianWins, dict)

    def test_yearStatSheet(self):
        from leeger.util.stat_sheet import yearStatSheet
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1, teamBScore=2)
        week = Week(weekNumber=1, matchups=[matchup])
        year = Year(yearNumber=2000, teams=teams, weeks=[week])

        yearStatSheet = yearStatSheet(year)

        self.assertIsInstance(yearStatSheet, YearStatSheet)
        self.assertIsInstance(yearStatSheet.gamesPlayed, dict)
        self.assertIsInstance(yearStatSheet.wins, dict)
        self.assertIsInstance(yearStatSheet.losses, dict)
        self.assertIsInstance(yearStatSheet.ties, dict)
        self.assertIsInstance(yearStatSheet.winPercentage, dict)
        self.assertIsInstance(yearStatSheet.wal, dict)
        self.assertIsInstance(yearStatSheet.walPerGame, dict)

        self.assertIsInstance(yearStatSheet.awal, dict)
        self.assertIsInstance(yearStatSheet.awalPerGame, dict)
        self.assertIsInstance(yearStatSheet.opponentAWAL, dict)
        self.assertIsInstance(yearStatSheet.opponentAWALPerGame, dict)

        self.assertIsInstance(yearStatSheet.smartWins, dict)
        self.assertIsInstance(yearStatSheet.smartWinsPerGame, dict)
        self.assertIsInstance(yearStatSheet.opponentSmartWins, dict)
        self.assertIsInstance(yearStatSheet.opponentSmartWinsPerGame, dict)

        self.assertIsInstance(yearStatSheet.pointsScored, dict)
        self.assertIsInstance(yearStatSheet.pointsScoredPerGame, dict)
        self.assertIsInstance(yearStatSheet.opponentPointsScored, dict)
        self.assertIsInstance(yearStatSheet.opponentPointsScoredPerGame, dict)

        self.assertIsInstance(yearStatSheet.scoringShare, dict)
        self.assertIsInstance(yearStatSheet.opponentScoringShare, dict)
        self.assertIsInstance(yearStatSheet.maxScoringShare, dict)
        self.assertIsInstance(yearStatSheet.minScoringShare, dict)

        self.assertIsInstance(yearStatSheet.maxScore, dict)
        self.assertIsInstance(yearStatSheet.minScore, dict)

        self.assertIsInstance(yearStatSheet.scoringStandardDeviation, dict)

        self.assertIsInstance(yearStatSheet.plusMinus, dict)

        self.assertIsInstance(yearStatSheet.teamScore, dict)
        self.assertIsInstance(yearStatSheet.teamSuccess, dict)
        self.assertIsInstance(yearStatSheet.teamLuck, dict)

        # check optional stats are None
        self.assertIsNone(yearStatSheet.leagueMedianWins)
        self.assertIsNone(yearStatSheet.totalGames)
        self.assertIsNone(yearStatSheet.opponentLeagueMedianWins)

        self.assertIsNone(yearStatSheet.ownerNames)
        self.assertIsNone(yearStatSheet.years)

    def test_yearStatSheet_ownerNamesAndYearsGiven(self):
        from leeger.util.stat_sheet import yearStatSheet
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1, teamBScore=2)
        week = Week(weekNumber=1, matchups=[matchup])
        year = Year(yearNumber=2000, teams=teams, weeks=[week])

        ownerNames = {
            teams[0].id: "owner0",
            teams[1].id: "owner1"
        }

        years = {
            teams[0].id: 2000,
            teams[1].id: 2000
        }

        yearStatSheet = yearStatSheet(year, ownerNames=ownerNames, years=years)

        self.assertIsInstance(yearStatSheet, YearStatSheet)
        self.assertIsInstance(yearStatSheet.gamesPlayed, dict)
        self.assertIsInstance(yearStatSheet.wins, dict)
        self.assertIsInstance(yearStatSheet.losses, dict)
        self.assertIsInstance(yearStatSheet.ties, dict)
        self.assertIsInstance(yearStatSheet.winPercentage, dict)
        self.assertIsInstance(yearStatSheet.wal, dict)
        self.assertIsInstance(yearStatSheet.walPerGame, dict)

        self.assertIsInstance(yearStatSheet.awal, dict)
        self.assertIsInstance(yearStatSheet.awalPerGame, dict)
        self.assertIsInstance(yearStatSheet.opponentAWAL, dict)
        self.assertIsInstance(yearStatSheet.opponentAWALPerGame, dict)

        self.assertIsInstance(yearStatSheet.smartWins, dict)
        self.assertIsInstance(yearStatSheet.smartWinsPerGame, dict)
        self.assertIsInstance(yearStatSheet.opponentSmartWins, dict)
        self.assertIsInstance(yearStatSheet.opponentSmartWinsPerGame, dict)

        self.assertIsInstance(yearStatSheet.pointsScored, dict)
        self.assertIsInstance(yearStatSheet.pointsScoredPerGame, dict)
        self.assertIsInstance(yearStatSheet.opponentPointsScored, dict)
        self.assertIsInstance(yearStatSheet.opponentPointsScoredPerGame, dict)

        self.assertIsInstance(yearStatSheet.scoringShare, dict)
        self.assertIsInstance(yearStatSheet.opponentScoringShare, dict)
        self.assertIsInstance(yearStatSheet.maxScoringShare, dict)
        self.assertIsInstance(yearStatSheet.minScoringShare, dict)

        self.assertIsInstance(yearStatSheet.maxScore, dict)
        self.assertIsInstance(yearStatSheet.minScore, dict)

        self.assertIsInstance(yearStatSheet.scoringStandardDeviation, dict)

        self.assertIsInstance(yearStatSheet.plusMinus, dict)

        self.assertIsInstance(yearStatSheet.teamScore, dict)
        self.assertIsInstance(yearStatSheet.teamSuccess, dict)
        self.assertIsInstance(yearStatSheet.teamLuck, dict)

        self.assertIsInstance(yearStatSheet.ownerNames, dict)
        self.assertIsInstance(yearStatSheet.years, dict)

        # check optional stats are None
        self.assertIsNone(yearStatSheet.leagueMedianWins)
        self.assertIsNone(yearStatSheet.totalGames)
        self.assertIsNone(yearStatSheet.opponentLeagueMedianWins)

    def test_yearStatSheet_leagueMedianGamesIsTrue(self):
        from leeger.util.stat_sheet import yearStatSheet
        owners, teams = getNDefaultOwnersAndTeams(2)

        matchup = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1, teamBScore=2)
        week = Week(weekNumber=1, matchups=[matchup])
        yearSettings = YearSettings(leagueMedianGames=True)
        year = Year(yearNumber=2000, teams=teams, weeks=[week], yearSettings=yearSettings)

        yearStatSheet = yearStatSheet(year)

        self.assertIsInstance(yearStatSheet, YearStatSheet)
        self.assertIsInstance(yearStatSheet.leagueMedianWins, dict)
        self.assertIsInstance(yearStatSheet.totalGames, dict)
        self.assertIsInstance(yearStatSheet.opponentLeagueMedianWins, dict)



import math
import os
import tempfile
import unittest

from openpyxl import load_workbook
from openpyxl.utils import get_column_letter

from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Team import Team
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year
from leeger.util.excel import leagueToExcel, yearToExcel
from leeger.util.stat_sheet import yearStatSheet
from test.helper.prototypes import getNDefaultOwnersAndTeams


class TestExcel(unittest.TestCase):
    def test_leagueToExcel_happyPath(self):
        owners, teams1 = getNDefaultOwnersAndTeams(2)
        teams1[0].name = "a"
        teams1[1].name = "b"
        matchup1 = Matchup(teamAId=teams1[0].id, teamBId=teams1[1].id, teamAScore=1, teamBScore=2)
        week1 = Week(weekNumber=1, matchups=[matchup1])
        year1 = Year(yearNumber=2000, teams=teams1, weeks=[week1])

        teams2 = [Team(ownerId=owners[0].id, name="a2"), Team(ownerId=owners[1].id, name="b2")]
        matchup2 = Matchup(teamAId=teams2[0].id, teamBId=teams2[1].id, teamAScore=1, teamBScore=2)
        week2 = Week(weekNumber=1, matchups=[matchup2])
        year2 = Year(yearNumber=2001, teams=teams2, weeks=[week2])

        teams3 = [Team(ownerId=owners[0].id, name="a3"), Team(ownerId=owners[1].id, name="b3")]
        matchup3 = Matchup(teamAId=teams3[0].id, teamBId=teams3[1].id, teamAScore=1, teamBScore=2)
        week3 = Week(weekNumber=1, matchups=[matchup3])
        year3 = Year(yearNumber=2002, teams=teams3, weeks=[week3])

        league = League(name="", owners=owners, years=[year1, year2, year3])

        with tempfile.TemporaryDirectory() as tempDir:
            fullPath = os.path.join(tempDir, "tmp.xlsx")
            leagueToExcel(league, fullPath)

            # open created Excel file to check that it was saved correctly
            workbook = load_workbook(filename=fullPath)

            # check sheets
            self.assertEqual(9, len(workbook.sheetnames))
            self.assertEqual("2000 Teams", workbook.sheetnames[0])
            self.assertEqual("2000 Matchups", workbook.sheetnames[1])
            self.assertEqual("2001 Teams", workbook.sheetnames[2])
            self.assertEqual("2001 Matchups", workbook.sheetnames[3])
            self.assertEqual("2002 Teams", workbook.sheetnames[4])
            self.assertEqual("2002 Matchups", workbook.sheetnames[5])
            self.assertEqual("All Time Teams", workbook.sheetnames[6])
            self.assertEqual("All Time Matchups", workbook.sheetnames[7])
            self.assertEqual("All Time Owners", workbook.sheetnames[8])

            # check sheet values
            # test year worksheet values
            worksheet2000TeamValues = [
                ["Team", "Games Played", "Wins", "Losses", "Ties", "Win Percentage", "WAL",
                 "WAL Per Game", "AWAL", "AWAL Per Game", "Opponent AWAL", "Opponent AWAL Per Game", "Smart Wins",
                 "Smart Wins Per Game", "Opponent Smart Wins", "Opponent Smart Wins Per Game", "Points Scored",
                 "Points Scored Per Game", "Opponent Points Scored", "Opponent Points Scored Per Game", "Scoring Share",
                 "Opponent Scoring Share", "Max Scoring Share", "Min Scoring Share", "Max Score", "Min Score",
                 "Scoring Standard Deviation", "Plus/Minus", "Team Score", "Team Success", "Team Luck"],
                ["a", 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 2, 2,
                 33.33333333333334, 66.66666666666667, 33.33333333333334, 33.33333333333334,
                 1, 1, 0, -1, 66.76666666666667, 66.76666666666667, 0],
                ["b", 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 2, 2, 1, 1,
                 66.66666666666667, 33.33333333333334, 66.66666666666667, 66.66666666666667,
                 2, 2, 0, 1, 233.5333333333333, 233.5333333333333, 0]]

            worksheet2000MatchupValues = [
                ["Team For", "Team Against", "Week Number", "Matchup Type", "Points For", "Points Against"],
                ["a", "b", 1, "REGULAR_SEASON", 1, 2],
                ["b", "a", 1, "REGULAR_SEASON", 2, 1]
            ]

            worksheet2001TeamValues = [
                ["Team", "Games Played", "Wins", "Losses", "Ties", "Win Percentage", "WAL",
                 "WAL Per Game", "AWAL", "AWAL Per Game", "Opponent AWAL", "Opponent AWAL Per Game", "Smart Wins",
                 "Smart Wins Per Game", "Opponent Smart Wins", "Opponent Smart Wins Per Game", "Points Scored",
                 "Points Scored Per Game", "Opponent Points Scored", "Opponent Points Scored Per Game", "Scoring Share",
                 "Opponent Scoring Share", "Max Scoring Share", "Min Scoring Share", "Max Score", "Min Score",
                 "Scoring Standard Deviation", "Plus/Minus", "Team Score", "Team Success", "Team Luck"],
                ["a2", 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 2, 2,
                 33.33333333333334, 66.66666666666667, 33.33333333333334, 33.33333333333334,
                 1, 1, 0, -1, 66.76666666666667, 66.76666666666667, 0],
                ["b2", 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 2, 2, 1, 1,
                 66.66666666666667, 33.33333333333334, 66.66666666666667, 66.66666666666667,
                 2, 2, 0, 1, 233.5333333333333, 233.5333333333333, 0]]

            worksheet2001MatchupValues = [
                ["Team For", "Team Against", "Week Number", "Matchup Type", "Points For", "Points Against"],
                ["a2", "b2", 1, "REGULAR_SEASON", 1, 2],
                ["b2", "a2", 1, "REGULAR_SEASON", 2, 1]
            ]

            worksheet2002TeamValues = [
                ["Team", "Games Played", "Wins", "Losses", "Ties", "Win Percentage", "WAL",
                 "WAL Per Game", "AWAL", "AWAL Per Game", "Opponent AWAL", "Opponent AWAL Per Game", "Smart Wins",
                 "Smart Wins Per Game", "Opponent Smart Wins", "Opponent Smart Wins Per Game", "Points Scored",
                 "Points Scored Per Game", "Opponent Points Scored", "Opponent Points Scored Per Game", "Scoring Share",
                 "Opponent Scoring Share", "Max Scoring Share", "Min Scoring Share", "Max Score", "Min Score",
                 "Scoring Standard Deviation", "Plus/Minus", "Team Score", "Team Success", "Team Luck"],
                ["a3", 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 2, 2,
                 33.33333333333334, 66.66666666666667, 33.33333333333334, 33.33333333333334,
                 1, 1, 0, -1, 66.76666666666667, 66.76666666666667, 0],
                ["b3", 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 2, 2, 1, 1,
                 66.66666666666667, 33.33333333333334, 66.66666666666667, 66.66666666666667,
                 2, 2, 0, 1, 233.5333333333333, 233.5333333333333, 0]]

            worksheet2002MatchupValues = [
                ["Team For", "Team Against", "Week Number", "Matchup Type", "Points For", "Points Against"],
                ["a3", "b3", 1, "REGULAR_SEASON", 1, 2],
                ["b3", "a3", 1, "REGULAR_SEASON", 2, 1]
            ]

            yearWorksheetsAndValues = [(workbook["2000 Teams"], worksheet2000TeamValues),
                                       (workbook["2000 Matchups"], worksheet2000MatchupValues),
                                       (workbook["2001 Teams"], worksheet2001TeamValues),
                                       (workbook["2001 Matchups"], worksheet2001MatchupValues),
                                       (workbook["2002 Teams"], worksheet2002TeamValues),
                                       (workbook["2002 Matchups"], worksheet2002MatchupValues)]

            for i, (currentWorksheet, allWorksheetValues) in enumerate(yearWorksheetsAndValues):
                for j, worksheetValues in enumerate(allWorksheetValues):
                    for k, worksheetValue in enumerate(worksheetValues):
                        row = j + 1
                        column = get_column_letter(k + 1)
                        self.assertEqual(worksheetValue, currentWorksheet[f"{column}{row}"].value)
                # check that "next" cell is empty
                self.assertIsNone(currentWorksheet["A4"].value)
                # check legend
                self.assertEqual("Filters Applied", currentWorksheet["A6"].value)
                self.assertEqual("Week Number Start: 1", currentWorksheet["A7"].value)
                self.assertEqual("Week Number End: 1", currentWorksheet["A8"].value)
                self.assertEqual("Only Regular Season: False", currentWorksheet["A9"].value)
                self.assertEqual("Only Post Season: False", currentWorksheet["A10"].value)
                self.assertEqual("Only Championship: False", currentWorksheet["A11"].value)
                self.assertEqual("Include Multi-Week Matchups: True", currentWorksheet["A12"].value)
                self.assertIsNone(currentWorksheet["A13"].value)

            # test all time teams worksheet values
            allTimeTeamsWorksheet = workbook["All Time Teams"]

            allTimeTeamsValues = [
                ["Team", "Owner", "Year", "Games Played", "Wins", "Losses", "Ties", "Win Percentage", "WAL",
                 "WAL Per Game", "AWAL", "AWAL Per Game", "Opponent AWAL", "Opponent AWAL Per Game", "Smart Wins",
                 "Smart Wins Per Game", "Opponent Smart Wins", "Opponent Smart Wins Per Game", "Points Scored",
                 "Points Scored Per Game", "Opponent Points Scored", "Opponent Points Scored Per Game", "Scoring Share",
                 "Opponent Scoring Share", "Max Scoring Share", "Min Scoring Share", "Max Score", "Min Score",
                 "Scoring Standard Deviation", "Plus/Minus", "Team Score", "Team Success", "Team Luck"],
                ["a", "1", 2000, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 2, 2,
                 33.33333333333334, 66.66666666666667, 33.33333333333334, 33.33333333333334,
                 1, 1, 0, -1, 66.76666666666667, 66.76666666666667, 0],
                ["b", "2", 2000, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 2, 2, 1, 1,
                 66.66666666666667, 33.33333333333334, 66.66666666666667, 66.66666666666667,
                 2, 2, 0, 1, 233.5333333333333, 233.5333333333333, 0],
                ["a2", "1", 2001, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 2, 2,
                 33.33333333333334, 66.66666666666667, 33.33333333333334, 33.33333333333334,
                 1, 1, 0, -1, 66.76666666666667, 66.76666666666667, 0],
                ["b2", "2", 2001, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 2, 2, 1, 1,
                 66.66666666666667, 33.33333333333334, 66.66666666666667, 66.66666666666667,
                 2, 2, 0, 1, 233.5333333333333, 233.5333333333333, 0],
                ["a3", "1", 2002, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 2, 2,
                 33.33333333333334, 66.66666666666667, 33.33333333333334, 33.33333333333334,
                 1, 1, 0, -1, 66.76666666666667, 66.76666666666667, 0],
                ["b3", "2", 2002, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 2, 2, 1, 1,
                 66.66666666666667, 33.33333333333334, 66.66666666666667, 66.66666666666667,
                 2, 2, 0, 1, 233.5333333333333, 233.5333333333333, 0]
            ]

            for rowNumber in range(1, 8):
                values = allTimeTeamsValues[rowNumber - 1]
                for columnNumber, value in enumerate(values):
                    cell = f"{get_column_letter(columnNumber + 1)}{rowNumber}"
                    self.assertEqual(values[columnNumber], allTimeTeamsWorksheet[cell].value)
            # check that "next" cell is empty
            self.assertIsNone(allTimeTeamsWorksheet["A8"].value)
            # check legend
            self.assertEqual("Filters Applied", allTimeTeamsWorksheet["A10"].value)
            self.assertEqual("Week Number Start: 1", allTimeTeamsWorksheet["A11"].value)
            self.assertEqual("Year Number Start: 2000", allTimeTeamsWorksheet["A12"].value)
            self.assertEqual("Week Number End: 1", allTimeTeamsWorksheet["A13"].value)
            self.assertEqual("Year Number End: 2002", allTimeTeamsWorksheet["A14"].value)
            self.assertEqual("Only Regular Season: False", allTimeTeamsWorksheet["A15"].value)
            self.assertEqual("Only Post Season: False", allTimeTeamsWorksheet["A16"].value)
            self.assertEqual("Only Championship: False", allTimeTeamsWorksheet["A17"].value)
            self.assertIsNone(allTimeTeamsWorksheet["A18"].value)

            # test all time matchups worksheet values
            allTimeMatchupsWorksheet = workbook["All Time Matchups"]

            allTimeMatchupsValues = [
                ["Team For", "Owner For", "Team Against", "Owner Against", "Year", "Week Number", "Matchup Type",
                 "Points For", "Points Against"],
                ["a", "1", "b", "2", 2000, 1, "REGULAR_SEASON", 1, 2],
                ["b", "2", "a", "1", 2000, 1, "REGULAR_SEASON", 2, 1],
                ["a2", "1", "b2", "2", 2001, 1, "REGULAR_SEASON", 1, 2],
                ["b2", "2", "a2", "1", 2001, 1, "REGULAR_SEASON", 2, 1],
                ["a3", "1", "b3", "2", 2002, 1, "REGULAR_SEASON", 1, 2],
                ["b3", "2", "a3", "1", 2002, 1, "REGULAR_SEASON", 2, 1]
            ]

            for rowNumber in range(1, 8):
                values = allTimeMatchupsValues[rowNumber - 1]
                for columnNumber, value in enumerate(values):
                    cell = f"{get_column_letter(columnNumber + 1)}{rowNumber}"
                    self.assertEqual(values[columnNumber], allTimeMatchupsWorksheet[cell].value)
            # check that "next" cell is empty
            self.assertIsNone(allTimeMatchupsWorksheet["A8"].value)
            # check legend
            self.assertEqual("Filters Applied", allTimeMatchupsWorksheet["A10"].value)
            self.assertEqual("Week Number Start: 1", allTimeMatchupsWorksheet["A11"].value)
            self.assertEqual("Year Number Start: 2000", allTimeMatchupsWorksheet["A12"].value)
            self.assertEqual("Week Number End: 1", allTimeMatchupsWorksheet["A13"].value)
            self.assertEqual("Year Number End: 2002", allTimeMatchupsWorksheet["A14"].value)
            self.assertEqual("Only Regular Season: False", allTimeMatchupsWorksheet["A15"].value)
            self.assertEqual("Only Post Season: False", allTimeMatchupsWorksheet["A16"].value)
            self.assertEqual("Only Championship: False", allTimeMatchupsWorksheet["A17"].value)
            self.assertIsNone(allTimeMatchupsWorksheet["A18"].value)

            # test all time owners worksheet values
            allTimeOwnersWorksheet = workbook["All Time Owners"]

            allTimeOwnersValues = [
                ["Owner", "Games Played", "Wins", "Losses", "Ties", "Win Percentage", "WAL", "WAL Per Game",
                 "AWAL", "AWAL Per Game", "Opponent AWAL", "Opponent AWAL Per Game", "Smart Wins",
                 "Smart Wins Per Game", "Opponent Smart Wins", "Opponent Smart Wins Per Game", "Points Scored",
                 "Points Scored Per Game", "Opponent Points Scored", "Opponent Points Scored Per Game", "Scoring Share",
                 "Opponent Scoring Share", "Max Scoring Share", "Min Scoring Share", "Max Score", "Min Score",
                 "Scoring Standard Deviation", "Plus/Minus", "Adjusted Team Score", "Adjusted Team Success",
                 "Adjusted Team Luck"],
                ["1", 3, 0, 3, 0, 0, 0, 0, 0, 0, 3, 1, 0.6, 0.2, 2.4, 0.8, 3, 1, 6, 2,
                 33.33333333333334, 66.66666666666667, 33.33333333333334, 33.33333333333334,
                 1, 1, 0, -3, 66.76666666666667, 66.76666666666667, 0],
                ["2", 3, 3, 0, 0, 1, 3, 1, 3, 1, 0, 0, 2.4, 0.8, 0.6, 0.2, 6, 2, 3, 1,
                 66.66666666666667, 33.33333333333334, 66.66666666666667, 66.66666666666667,
                 2, 2, 0, 3, 233.5333333333333, 233.5333333333333, 0]
            ]

            for rowNumber in range(1, 4):
                values = allTimeOwnersValues[rowNumber - 1]
                for columnNumber, value in enumerate(values):
                    cell = f"{get_column_letter(columnNumber + 1)}{rowNumber}"
                    self.assertEqual(values[columnNumber], allTimeOwnersWorksheet[cell].value)
            # check that "next" cell is empty
            self.assertIsNone(allTimeOwnersWorksheet["A4"].value)
            # check legend
            self.assertEqual("Filters Applied", allTimeOwnersWorksheet["A6"].value)
            self.assertEqual("Week Number Start: 1", allTimeOwnersWorksheet["A7"].value)
            self.assertEqual("Year Number Start: 2000", allTimeOwnersWorksheet["A8"].value)
            self.assertEqual("Week Number End: 1", allTimeOwnersWorksheet["A9"].value)
            self.assertEqual("Year Number End: 2002", allTimeOwnersWorksheet["A10"].value)
            self.assertEqual("Only Regular Season: False", allTimeOwnersWorksheet["A11"].value)
            self.assertEqual("Only Post Season: False", allTimeOwnersWorksheet["A12"].value)
            self.assertEqual("Only Championship: False", allTimeOwnersWorksheet["A13"].value)
            self.assertIsNone(allTimeOwnersWorksheet["A14"].value)

    def test_leagueToExcel_leagueIsNone_raisesException(self):

        with self.assertRaises(ValueError) as context:
            leagueToExcel(None, "")
        self.assertEqual("'league' has not been set.", str(context.exception))

    def test_leagueToExcel_fileAlreadyExists_raisesException(self):
        owners, teams1 = getNDefaultOwnersAndTeams(2)
        teams1[0].name = "a"
        teams1[1].name = "b"
        matchup1 = Matchup(teamAId=teams1[0].id, teamBId=teams1[1].id, teamAScore=1, teamBScore=2)
        week1 = Week(weekNumber=1, matchups=[matchup1])
        year1 = Year(yearNumber=2000, teams=teams1, weeks=[week1])

        teams2 = [Team(ownerId=owners[0].id, name="a2"), Team(ownerId=owners[1].id, name="b2")]
        matchup2 = Matchup(teamAId=teams2[0].id, teamBId=teams2[1].id, teamAScore=1, teamBScore=2)
        week2 = Week(weekNumber=1, matchups=[matchup2])
        year2 = Year(yearNumber=2001, teams=teams2, weeks=[week2])

        teams3 = [Team(ownerId=owners[0].id, name="a3"), Team(ownerId=owners[1].id, name="b3")]
        matchup3 = Matchup(teamAId=teams3[0].id, teamBId=teams3[1].id, teamAScore=1, teamBScore=2)
        week3 = Week(weekNumber=1, matchups=[matchup3])
        year3 = Year(yearNumber=1999, teams=teams3, weeks=[week3])

        league = League(name="", owners=owners, years=[year1, year2, year3])

        with self.assertRaises(FileExistsError) as context:
            with tempfile.TemporaryDirectory() as tempDir:
                fullPath = os.path.join(tempDir, "tmp.xlsx")
                # create a file with this same name/path first
                with open(fullPath, "w") as f:
                    ...
                leagueToExcel(league, fullPath)
        self.assertEqual(f"Cannot create file at path: '{fullPath}' because there is already a file there.",
                         str(context.exception))

    def test_yearToExcel_excelSheetDoesntAlreadyExist(self):
        owners, teams = getNDefaultOwnersAndTeams(2)
        teams[0].name = "a"
        teams[1].name = "b"
        matchup = Matchup(teamAId=teams[0].id, teamBId=teams[1].id, teamAScore=1, teamBScore=2)
        week = Week(weekNumber=1, matchups=[matchup])
        year = Year(yearNumber=2000, teams=teams, weeks=[week])

        with tempfile.TemporaryDirectory() as tempDir:
            fullPath = os.path.join(tempDir, "tmp.xlsx")
            yearToExcel(year, fullPath)

            # open created Excel file to check that it was saved correctly
            workbook = load_workbook(filename=fullPath)
            self.assertEqual(2, len(workbook.sheetnames))
            self.assertEqual("2000 Teams", workbook.sheetnames[0])
            self.assertEqual("2000 Matchups", workbook.sheetnames[1])

            # test team worksheet values
            worksheet = workbook["2000 Teams"]

            worksheetTeamValues = [
                ["Team", "Games Played", "Wins", "Losses", "Ties", "Win Percentage", "WAL",
                 "WAL Per Game", "AWAL", "AWAL Per Game", "Opponent AWAL", "Opponent AWAL Per Game", "Smart Wins",
                 "Smart Wins Per Game", "Opponent Smart Wins", "Opponent Smart Wins Per Game", "Points Scored",
                 "Points Scored Per Game", "Opponent Points Scored", "Opponent Points Scored Per Game", "Scoring Share",
                 "Opponent Scoring Share", "Max Scoring Share", "Min Scoring Share", "Max Score", "Min Score",
                 "Scoring Standard Deviation", "Plus/Minus", "Team Score", "Team Success", "Team Luck"],
                ["a", 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 2, 2,
                 33.33333333333334, 66.66666666666667, 33.33333333333334, 33.33333333333334,
                 1, 1, 0, -1, 66.76666666666667, 66.76666666666667, 0],
                ["b", 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 2, 2, 1, 1,
                 66.66666666666667, 33.33333333333334, 66.66666666666667, 66.66666666666667,
                 2, 2, 0, 1, 233.5333333333333, 233.5333333333333, 0]]

            for rowNumber in range(1, 4):
                values = worksheetTeamValues[rowNumber - 1]
                for columnNumber, value in enumerate(values):
                    cell = f"{get_column_letter(columnNumber + 1)}{rowNumber}"
                    self.assertEqual(values[columnNumber], worksheet[cell].value)
            # check that "next" cell is empty
            self.assertIsNone(worksheet["A4"].value)

            # check legend
            self.assertEqual("Filters Applied", worksheet["A6"].value)
            self.assertEqual("Week Number Start: 1", worksheet["A7"].value)
            self.assertEqual("Week Number End: 1", worksheet["A8"].value)
            self.assertEqual("Only Regular Season: False", worksheet["A9"].value)
            self.assertEqual("Only Post Season: False", worksheet["A10"].value)
            self.assertEqual("Only Championship: False", worksheet["A11"].value)
            self.assertEqual("Include Multi-Week Matchups: True", worksheet["A12"].value)
            self.assertIsNone(worksheet["A13"].value)

            # make sure stats were transferred properly
            statsWithTitles = yearStatSheet(year).preferredOrderWithTitle()
            for row, teamId in enumerate([team.id for team in year.teams]):
                for col, statWithTitle in enumerate(statsWithTitles):
                    char = get_column_letter(col + 2)
                    if row == 1:
                        # check stat header
                        self.assertEqual(statWithTitle[0], worksheet[f"{char}{row}"].value)
                    # check stat value
                    # due to Excel rounding values, we assert that the values are very, very close
                    assert math.isclose(float(statWithTitle[1][teamId]), worksheet[f"{char}{row + 2}"].value,
                                        rel_tol=0.000000000000001)

            # test matchup worksheet values
            worksheet = workbook["2000 Matchups"]

            worksheetMatchupValues = [
                ["Team For", "Team Against", "Week Number", "Matchup Type", "Points For", "Points Against"],
                ["a", "b", 1, "REGULAR_SEASON", 1, 2],
                ["b", "a", 1, "REGULAR_SEASON", 2, 1]
            ]
            for rowNumber in range(1, 4):
                values = worksheetMatchupValues[rowNumber - 1]
                for columnNumber, value in enumerate(values):
                    cell = f"{get_column_letter(columnNumber + 1)}{rowNumber}"
                    self.assertEqual(values[columnNumber], worksheet[cell].value)
            # check that "next" cell is empty
            self.assertIsNone(worksheet["A4"].value)

            # check legend
            self.assertEqual("Filters Applied", worksheet["A6"].value)
            self.assertEqual("Week Number Start: 1", worksheet["A7"].value)
            self.assertEqual("Week Number End: 1", worksheet["A8"].value)
            self.assertEqual("Only Regular Season: False", worksheet["A9"].value)
            self.assertEqual("Only Post Season: False", worksheet["A10"].value)
            self.assertEqual("Only Championship: False", worksheet["A11"].value)
            self.assertEqual("Include Multi-Week Matchups: True", worksheet["A12"].value)
            self.assertIsNone(worksheet["A13"].value)

    def test_yearToExcel_excelSheetAlreadyExists(self):
        owners, teams1 = getNDefaultOwnersAndTeams(2)
        teams1[0].name = "a1"
        teams1[1].name = "b1"
        matchup1 = Matchup(teamAId=teams1[0].id, teamBId=teams1[1].id, teamAScore=1, teamBScore=2)
        week1 = Week(weekNumber=1, matchups=[matchup1])
        year1 = Year(yearNumber=2000, teams=teams1, weeks=[week1])

        teams2 = [Team(ownerId=owners[0].id, name="a2"), Team(ownerId=owners[1].id, name="b2")]
        matchup2 = Matchup(teamAId=teams2[0].id, teamBId=teams2[1].id, teamAScore=1, teamBScore=2)
        week2 = Week(weekNumber=1, matchups=[matchup2])
        year2 = Year(yearNumber=2001, teams=teams2, weeks=[week2])

        with tempfile.TemporaryDirectory() as tempDir:
            fullPath = os.path.join(tempDir, "tmp.xlsx")
            yearToExcel(year1, fullPath)
            # "new" year
            yearToExcel(year2, fullPath)

            # open created Excel file to check that it was saved correctly
            workbook = load_workbook(filename=fullPath)
            self.assertEqual(4, len(workbook.sheetnames))
            self.assertEqual("2000 Teams", workbook.sheetnames[0])
            self.assertEqual("2000 Matchups", workbook.sheetnames[1])
            self.assertEqual("2001 Teams", workbook.sheetnames[2])
            self.assertEqual("2001 Matchups", workbook.sheetnames[3])

            # test team worksheet values
            worksheet2000 = workbook["2000 Teams"]
            worksheet2000Values = [
                ["Team", "Games Played", "Wins", "Losses", "Ties", "Win Percentage", "WAL",
                 "WAL Per Game", "AWAL", "AWAL Per Game", "Opponent AWAL", "Opponent AWAL Per Game", "Smart Wins",
                 "Smart Wins Per Game", "Opponent Smart Wins", "Opponent Smart Wins Per Game", "Points Scored",
                 "Points Scored Per Game", "Opponent Points Scored", "Opponent Points Scored Per Game", "Scoring Share",
                 "Opponent Scoring Share", "Max Scoring Share", "Min Scoring Share", "Max Score", "Min Score",
                 "Scoring Standard Deviation", "Plus/Minus", "Team Score", "Team Success", "Team Luck"],
                ["a1", 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 2, 2,
                 33.33333333333334, 66.66666666666667, 33.33333333333334, 33.33333333333334,
                 1, 1, 0, -1, 66.76666666666667, 66.76666666666667, 0],
                ["b1", 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 2, 2, 1, 1,
                 66.66666666666667, 33.33333333333334, 66.66666666666667, 66.66666666666667,
                 2, 2, 0, 1, 233.5333333333333, 233.5333333333333, 0]]

            for rowNumber in range(1, 4):
                values = worksheet2000Values[rowNumber - 1]
                for columnNumber, value in enumerate(values):
                    cell = f"{get_column_letter(columnNumber + 1)}{rowNumber}"
                    self.assertEqual(values[columnNumber], worksheet2000[cell].value)
            # check that "next" cell is empty
            self.assertIsNone(worksheet2000["A4"].value)
            # check legend
            self.assertEqual("Filters Applied", worksheet2000["A6"].value)
            self.assertEqual("Week Number Start: 1", worksheet2000["A7"].value)
            self.assertEqual("Week Number End: 1", worksheet2000["A8"].value)
            self.assertEqual("Only Regular Season: False", worksheet2000["A9"].value)
            self.assertEqual("Only Post Season: False", worksheet2000["A10"].value)
            self.assertEqual("Only Championship: False", worksheet2000["A11"].value)
            self.assertEqual("Include Multi-Week Matchups: True", worksheet2000["A12"].value)
            self.assertIsNone(worksheet2000["A13"].value)

            # test matchup worksheet values
            worksheet = workbook["2000 Matchups"]

            worksheet2000MatchupValues = [
                ["Team For", "Team Against", "Week Number", "Matchup Type", "Points For", "Points Against"],
                ["a1", "b1", 1, "REGULAR_SEASON", 1, 2],
                ["b1", "a1", 1, "REGULAR_SEASON", 2, 1]
            ]
            for rowNumber in range(1, 4):
                values = worksheet2000MatchupValues[rowNumber - 1]
                for columnNumber, value in enumerate(values):
                    cell = f"{get_column_letter(columnNumber + 1)}{rowNumber}"
                    self.assertEqual(values[columnNumber], worksheet[cell].value)
            # check that "next" cell is empty
            self.assertIsNone(worksheet["A4"].value)

            worksheet2001 = workbook["2001 Teams"]
            worksheet2001Values = [
                ["Team", "Games Played", "Wins", "Losses", "Ties", "Win Percentage", "WAL",
                 "WAL Per Game", "AWAL", "AWAL Per Game", "Opponent AWAL", "Opponent AWAL Per Game", "Smart Wins",
                 "Smart Wins Per Game", "Opponent Smart Wins", "Opponent Smart Wins Per Game", "Points Scored",
                 "Points Scored Per Game", "Opponent Points Scored", "Opponent Points Scored Per Game", "Scoring Share",
                 "Opponent Scoring Share", "Max Scoring Share", "Min Scoring Share", "Max Score", "Min Score",
                 "Scoring Standard Deviation", "Plus/Minus", "Team Score", "Team Success", "Team Luck"],
                ["a2", 1, 0, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 2, 2,
                 33.33333333333334, 66.66666666666667, 33.33333333333334, 33.33333333333334,
                 1, 1, 0, -1, 66.76666666666667, 66.76666666666667, 0],
                ["b2", 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 2, 2, 1, 1,
                 66.66666666666667, 33.33333333333334, 66.66666666666667, 66.66666666666667,
                 2, 2, 0, 1, 233.5333333333333, 233.5333333333333, 0]]

            for rowNumber in range(1, 4):
                values = worksheet2001Values[rowNumber - 1]
                for columnNumber, value in enumerate(values):
                    cell = f"{get_column_letter(columnNumber + 1)}{rowNumber}"
                    self.assertEqual(values[columnNumber], worksheet2001[cell].value)
            # check that "next" cell is empty
            self.assertIsNone(worksheet2001["A4"].value)

            # test matchup worksheet values
            worksheet = workbook["2001 Matchups"]

            worksheet2001MatchupValues = [
                ["Team For", "Team Against", "Week Number", "Matchup Type", "Points For", "Points Against"],
                ["a2", "b2", 1, "REGULAR_SEASON", 1, 2],
                ["b2", "a2", 1, "REGULAR_SEASON", 2, 1]
            ]
            for rowNumber in range(1, 4):
                values = worksheet2001MatchupValues[rowNumber - 1]
                for columnNumber, value in enumerate(values):
                    cell = f"{get_column_letter(columnNumber + 1)}{rowNumber}"
                    self.assertEqual(values[columnNumber], worksheet[cell].value)
            # check that "next" cell is empty
            self.assertIsNone(worksheet["A4"].value)

            # check values
            worksheet = workbook.worksheets[workbook.sheetnames.index("2000 Teams")]

            statsWithTitles = yearStatSheet(year1).preferredOrderWithTitle()
            for row, teamId in enumerate([team.id for team in year1.teams]):
                for col, statWithTitle in enumerate(statsWithTitles):
                    char = get_column_letter(col + 2)
                    if row == 1:
                        # check stat header
                        self.assertEqual(statWithTitle[0], worksheet[f"{char}{row}"].value)
                    # check stat value
                    # due to Excel rounding values, we assert that the values are very, very close
                    assert math.isclose(float(statWithTitle[1][teamId]), worksheet[f"{char}{row + 2}"].value,
                                        rel_tol=0.000000000000001)

            worksheet = workbook.worksheets[workbook.sheetnames.index("2001 Teams")]

            statsWithTitles = yearStatSheet(year2).preferredOrderWithTitle()
            for row, teamId in enumerate([team.id for team in year2.teams]):
                for col, statWithTitle in enumerate(statsWithTitles):
                    char = get_column_letter(col + 2)
                    if row == 1:
                        # check stat header
                        self.assertEqual(statWithTitle[0], worksheet[f"{char}{row}"].value)
                    # check stat value
                    # due to Excel rounding values, we assert that the values are very, very close
                    assert math.isclose(float(statWithTitle[1][teamId]), worksheet[f"{char}{row + 2}"].value,
                                        rel_tol=0.000000000000001)



import unittest

from leeger.calculator.all_time_calculator.SSLAllTimeCalculator import SSLAllTimeCalculator
from leeger.enum.MatchupType import MatchupType
from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year
from leeger.util.Deci import Deci
from test.helper.prototypes import getNDefaultOwnersAndTeams, getTeamsFromOwners


class TestSSLAllTimeCalculator(unittest.TestCase):


    def test_getAdjustedTeamScore_happyPath(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamScore(league)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamScore_noneIfNoGamesPlayed(self):
        owners, teamsA = getNDefaultOwnersAndTeams(3)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[1].id, teamBId=teamsA[2].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a])

        league = League(name="TEST", owners=owners, years=[yearA])
        response = SSLAllTimeCalculator.getAdjustedTeamScore(league, weekNumberEnd=1)

        self.assertIsInstance(response, dict)
        self.assertEqual(3, len(response.keys()))
        self.assertEqual(Deci("66.76666666666666666666666666"), response[owners[0].id])
        self.assertEqual(Deci("233.5333333333333333333333333"), response[owners[1].id])
        self.assertIsNone(response[owners[2].id])

    def test_getAdjustedTeamScore_onlyPostSeasonIsTrue(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        matchup3_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week3_a = Week(weekNumber=3, matchups=[matchup3_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a, week3_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamScore(league, onlyPostSeason=True)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamScore_onlyRegularSeasonIsTrue(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        matchup3_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week3_a = Week(weekNumber=3, matchups=[matchup3_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a, week3_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamScore(league, onlyRegularSeason=True)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamScore_onlyChampionshipIsTrue(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        matchup3_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_a = Week(weekNumber=3, matchups=[matchup3_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a, week3_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamScore(league, onlyChampionship=True)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("167.2166666666666666666666667"), response[owners[0].id])
        self.assertEqual(Deci("133.5333333333333333333333333"), response[owners[1].id])

    def test_getAdjustedTeamScore_yearNumberStartGivenWeekNumberStartGiven(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)
        teamsC = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        yearA = Year(yearNumber=1999, teams=teamsA, weeks=[week1_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        matchup2_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week2_b = Week(weekNumber=2, matchups=[matchup2_b])
        matchup3_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_b = Week(weekNumber=3, matchups=[matchup3_b])
        yearB = Year(yearNumber=2000, teams=teamsB, weeks=[week1_b, week2_b, week3_b])

        matchup1_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_c = Week(weekNumber=1, matchups=[matchup1_c])
        yearC = Year(yearNumber=2001, teams=teamsC, weeks=[week1_c])

        league = League(name="TEST", owners=owners, years=[yearA, yearB, yearC])

        response = SSLAllTimeCalculator.getAdjustedTeamScore(league, yearNumberStart=2000, weekNumberStart=2)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamScore_yearNumberEndGivenWeekNumberEndGiven(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)
        teamsC = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        yearA = Year(yearNumber=1999, teams=teamsA, weeks=[week1_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        matchup2_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week2_b = Week(weekNumber=2, matchups=[matchup2_b])
        matchup3_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_b = Week(weekNumber=3, matchups=[matchup3_b])
        yearB = Year(yearNumber=2000, teams=teamsB, weeks=[week1_b, week2_b, week3_b])

        matchup1_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_c = Week(weekNumber=1, matchups=[matchup1_c])
        yearC = Year(yearNumber=2001, teams=teamsC, weeks=[week1_c])

        league = League(name="TEST", owners=owners, years=[yearA, yearB, yearC])

        response = SSLAllTimeCalculator.getAdjustedTeamScore(league, yearNumberEnd=2000, weekNumberEnd=2)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamScore_yearNumberStartGivenWeekNumberStartGivenAndYearNumberEndGivenWeekNumberEndGiven(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)
        teamsC = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        yearA = Year(yearNumber=1999, teams=teamsA, weeks=[week1_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        matchup2_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week2_b = Week(weekNumber=2, matchups=[matchup2_b])
        matchup3_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_b = Week(weekNumber=3, matchups=[matchup3_b])
        yearB = Year(yearNumber=2000, teams=teamsB, weeks=[week1_b, week2_b, week3_b])

        matchup1_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_c = Week(weekNumber=1, matchups=[matchup1_c])
        matchup2_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week2_c = Week(weekNumber=2, matchups=[matchup2_c])
        yearC = Year(yearNumber=2001, teams=teamsC, weeks=[week1_c, week2_c])

        league = League(name="TEST", owners=owners, years=[yearA, yearB, yearC])

        response = SSLAllTimeCalculator.getAdjustedTeamScore(league, yearNumberStart=2000, weekNumberStart=2,
                                                             yearNumberEnd=2001, weekNumberEnd=1)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamSuccess_happyPath(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamSuccess(league)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamSuccess_noneIfNoGamesPlayed(self):
        owners, teamsA = getNDefaultOwnersAndTeams(3)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[1].id, teamBId=teamsA[2].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a])

        league = League(name="TEST", owners=owners, years=[yearA])
        response = SSLAllTimeCalculator.getAdjustedTeamSuccess(league, weekNumberEnd=1)

        self.assertIsInstance(response, dict)
        self.assertEqual(3, len(response.keys()))
        self.assertEqual(Deci("66.76666666666666666666666666"), response[owners[0].id])
        self.assertEqual(Deci("233.5333333333333333333333333"), response[owners[1].id])
        self.assertIsNone(response[owners[2].id])

    def test_getAdjustedTeamSuccess_onlyPostSeasonIsTrue(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        matchup3_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week3_a = Week(weekNumber=3, matchups=[matchup3_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a, week3_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamSuccess(league, onlyPostSeason=True)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamSuccess_onlyRegularSeasonIsTrue(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        matchup3_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week3_a = Week(weekNumber=3, matchups=[matchup3_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a, week3_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamSuccess(league, onlyRegularSeason=True)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamSuccess_onlyChampionshipIsTrue(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        matchup3_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_a = Week(weekNumber=3, matchups=[matchup3_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a, week3_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamSuccess(league, onlyChampionship=True)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("167.2166666666666666666666667"), response[owners[0].id])
        self.assertEqual(Deci("133.5333333333333333333333333"), response[owners[1].id])

    def test_getAdjustedTeamSuccess_yearNumberStartGivenWeekNumberStartGiven(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)
        teamsC = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        yearA = Year(yearNumber=1999, teams=teamsA, weeks=[week1_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        matchup2_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week2_b = Week(weekNumber=2, matchups=[matchup2_b])
        matchup3_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_b = Week(weekNumber=3, matchups=[matchup3_b])
        yearB = Year(yearNumber=2000, teams=teamsB, weeks=[week1_b, week2_b, week3_b])

        matchup1_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_c = Week(weekNumber=1, matchups=[matchup1_c])
        yearC = Year(yearNumber=2001, teams=teamsC, weeks=[week1_c])

        league = League(name="TEST", owners=owners, years=[yearA, yearB, yearC])

        response = SSLAllTimeCalculator.getAdjustedTeamSuccess(league, yearNumberStart=2000, weekNumberStart=2)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamSuccess_yearNumberEndGivenWeekNumberEndGiven(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)
        teamsC = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        yearA = Year(yearNumber=1999, teams=teamsA, weeks=[week1_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        matchup2_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week2_b = Week(weekNumber=2, matchups=[matchup2_b])
        matchup3_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_b = Week(weekNumber=3, matchups=[matchup3_b])
        yearB = Year(yearNumber=2000, teams=teamsB, weeks=[week1_b, week2_b, week3_b])

        matchup1_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_c = Week(weekNumber=1, matchups=[matchup1_c])
        yearC = Year(yearNumber=2001, teams=teamsC, weeks=[week1_c])

        league = League(name="TEST", owners=owners, years=[yearA, yearB, yearC])

        response = SSLAllTimeCalculator.getAdjustedTeamSuccess(league, yearNumberEnd=2000, weekNumberEnd=2)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamSuccess_yearNumberStartGivenWeekNumberStartGivenAndYearNumberEndGivenWeekNumberEndGiven(
            self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)
        teamsC = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        yearA = Year(yearNumber=1999, teams=teamsA, weeks=[week1_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        matchup2_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week2_b = Week(weekNumber=2, matchups=[matchup2_b])
        matchup3_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_b = Week(weekNumber=3, matchups=[matchup3_b])
        yearB = Year(yearNumber=2000, teams=teamsB, weeks=[week1_b, week2_b, week3_b])

        matchup1_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_c = Week(weekNumber=1, matchups=[matchup1_c])
        matchup2_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week2_c = Week(weekNumber=2, matchups=[matchup2_c])
        yearC = Year(yearNumber=2001, teams=teamsC, weeks=[week1_c, week2_c])

        league = League(name="TEST", owners=owners, years=[yearA, yearB, yearC])

        response = SSLAllTimeCalculator.getAdjustedTeamSuccess(league, yearNumberStart=2000, weekNumberStart=2,
                                                               yearNumberEnd=2001, weekNumberEnd=1)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("133.7333333333333199600000000"), response[owners[0].id])
        self.assertEqual(Deci("166.8666666666666499800000000"), response[owners[1].id])

    def test_getAdjustedTeamLuck_happyPath(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamLuck(league)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0"), response[owners[0].id])
        self.assertEqual(Deci("0"), response[owners[1].id])

    def test_getAdjustedTeamLuck_noneIfNoGamesPlayed(self):
        owners, teamsA = getNDefaultOwnersAndTeams(3)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[1].id, teamBId=teamsA[2].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a])

        league = League(name="TEST", owners=owners, years=[yearA])
        response = SSLAllTimeCalculator.getAdjustedTeamLuck(league, weekNumberEnd=1)

        self.assertIsInstance(response, dict)
        self.assertEqual(3, len(response.keys()))
        self.assertEqual(Deci("0"), response[owners[0].id])
        self.assertEqual(Deci("0"), response[owners[1].id])
        self.assertIsNone(response[owners[2].id])

    def test_getAdjustedTeamLuck_onlyPostSeasonIsTrue(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        matchup3_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week3_a = Week(weekNumber=3, matchups=[matchup3_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a, week3_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamLuck(league, onlyPostSeason=True)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0"), response[owners[0].id])
        self.assertEqual(Deci("0"), response[owners[1].id])

    def test_getAdjustedTeamLuck_onlyRegularSeasonIsTrue(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        matchup3_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week3_a = Week(weekNumber=3, matchups=[matchup3_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a, week3_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamLuck(league, onlyRegularSeason=True)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0"), response[owners[0].id])
        self.assertEqual(Deci("0"), response[owners[1].id])

    def test_getAdjustedTeamLuck_onlyChampionshipIsTrue(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        matchup2_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2)
        week2_a = Week(weekNumber=2, matchups=[matchup2_a])
        matchup3_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_a = Week(weekNumber=3, matchups=[matchup3_a])
        yearA = Year(yearNumber=2000, teams=teamsA, weeks=[week1_a, week2_a, week3_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        yearB = Year(yearNumber=2001, teams=teamsB, weeks=[week1_b])

        league = League(name="TEST", owners=owners, years=[yearA, yearB])

        response = SSLAllTimeCalculator.getAdjustedTeamLuck(league, onlyChampionship=True)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0"), response[owners[0].id])
        self.assertEqual(Deci("0"), response[owners[1].id])

    def test_getAdjustedTeamLuck_yearNumberStartGivenWeekNumberStartGiven(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)
        teamsC = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        yearA = Year(yearNumber=1999, teams=teamsA, weeks=[week1_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        matchup2_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week2_b = Week(weekNumber=2, matchups=[matchup2_b])
        matchup3_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_b = Week(weekNumber=3, matchups=[matchup3_b])
        yearB = Year(yearNumber=2000, teams=teamsB, weeks=[week1_b, week2_b, week3_b])

        matchup1_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_c = Week(weekNumber=1, matchups=[matchup1_c])
        yearC = Year(yearNumber=2001, teams=teamsC, weeks=[week1_c])

        league = League(name="TEST", owners=owners, years=[yearA, yearB, yearC])

        response = SSLAllTimeCalculator.getAdjustedTeamLuck(league, yearNumberStart=2000, weekNumberStart=2)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0"), response[owners[0].id])
        self.assertEqual(Deci("0"), response[owners[1].id])

    def test_getAdjustedTeamLuck_yearNumberEndGivenWeekNumberEndGiven(self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)
        teamsC = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        yearA = Year(yearNumber=1999, teams=teamsA, weeks=[week1_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        matchup2_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week2_b = Week(weekNumber=2, matchups=[matchup2_b])
        matchup3_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_b = Week(weekNumber=3, matchups=[matchup3_b])
        yearB = Year(yearNumber=2000, teams=teamsB, weeks=[week1_b, week2_b, week3_b])

        matchup1_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week1_c = Week(weekNumber=1, matchups=[matchup1_c])
        yearC = Year(yearNumber=2001, teams=teamsC, weeks=[week1_c])

        league = League(name="TEST", owners=owners, years=[yearA, yearB, yearC])

        response = SSLAllTimeCalculator.getAdjustedTeamLuck(league, yearNumberEnd=2000, weekNumberEnd=2)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0"), response[owners[0].id])
        self.assertEqual(Deci("0"), response[owners[1].id])

    def test_getAdjustedTeamLuck_yearNumberStartGivenWeekNumberStartGivenAndYearNumberEndGivenWeekNumberEndGiven(
            self):
        owners, teamsA = getNDefaultOwnersAndTeams(2)
        teamsB = getTeamsFromOwners(owners)
        teamsC = getTeamsFromOwners(owners)

        matchup1_a = Matchup(teamAId=teamsA[0].id, teamBId=teamsA[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_a = Week(weekNumber=1, matchups=[matchup1_a])
        yearA = Year(yearNumber=1999, teams=teamsA, weeks=[week1_a])

        matchup1_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week1_b = Week(weekNumber=1, matchups=[matchup1_b])
        matchup2_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2)
        week2_b = Week(weekNumber=2, matchups=[matchup2_b])
        matchup3_b = Matchup(teamAId=teamsB[0].id, teamBId=teamsB[1].id, teamAScore=1, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week3_b = Week(weekNumber=3, matchups=[matchup3_b])
        yearB = Year(yearNumber=2000, teams=teamsB, weeks=[week1_b, week2_b, week3_b])

        matchup1_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.PLAYOFF)
        week1_c = Week(weekNumber=1, matchups=[matchup1_c])
        matchup2_c = Matchup(teamAId=teamsC[0].id, teamBId=teamsC[1].id, teamAScore=10, teamBScore=2,
                             matchupType=MatchupType.CHAMPIONSHIP)
        week2_c = Week(weekNumber=2, matchups=[matchup2_c])
        yearC = Year(yearNumber=2001, teams=teamsC, weeks=[week1_c, week2_c])

        league = League(name="TEST", owners=owners, years=[yearA, yearB, yearC])

        response = SSLAllTimeCalculator.getAdjustedTeamLuck(league, yearNumberStart=2000, weekNumberStart=2,
                                                            yearNumberEnd=2001, weekNumberEnd=1)

        self.assertIsInstance(response, dict)
        self.assertEqual(2, len(response.keys()))
        self.assertEqual(Deci("0"), response[owners[0].id])
        self.assertEqual(Deci("0"), response[owners[1].id])



from leeger.model.filter import YearFilters, AllTimeFilters
from leeger.model.league import League, Year
from leeger.util.navigator import LeagueNavigator, YearNavigator
from leeger.util.stat_sheet import yearStatSheet


def allTimeTeamsStatSheet(league: League, **kwargs) -> list[tuple[str, dict]]:
    allTimeTeamsStatsWithTitles: list[tuple[str, dict]] = list()
    for year in league.years:
        ownerNames: dict[str, str] = dict()
        years: dict[str, int] = dict()
        teamIdToNameMap = dict()
        for team in year.teams:
            ownerNames[team.id] = LeagueNavigator.getOwnerById(league, team.ownerId).name
            years[team.id] = year.yearNumber
            teamIdToNameMap[team.id] = team.name
        yearStatsWithTitles = yearStatSheet(year,
                                            ownerNames=ownerNames,
                                            years=years,
                                            **kwargs).preferredOrderWithTitle()

        yearStatsWithTitles.insert(0, ("Team", teamIdToNameMap))
        allTimeTeamsStatsWithTitles += yearStatsWithTitles

    # condense stats with titles so there's only 1 list value for each title
    condensedAllTimeTeamsStatsWithTitles: list[tuple[str, dict]] = list()
    for titleStr, statsDict in allTimeTeamsStatsWithTitles:
        allTitlesInCondensedList = list()
        if len(condensedAllTimeTeamsStatsWithTitles) > 0:
            allTitlesInCondensedList = [values[0] for values in condensedAllTimeTeamsStatsWithTitles]
        if titleStr in allTitlesInCondensedList:
            # add to stats dict for the existing title
            for i, (title_s, stats_d) in enumerate(condensedAllTimeTeamsStatsWithTitles):
                if title_s == titleStr:
                    condensedAllTimeTeamsStatsWithTitles[i] = (title_s, stats_d | statsDict)
        else:
            condensedAllTimeTeamsStatsWithTitles.append((titleStr, statsDict))
    return condensedAllTimeTeamsStatsWithTitles


def yearMatchupsStatSheet(year: Year, includeOwnerIds: bool = False, includeYears: bool = False, **kwargs) -> tuple[
    list[tuple[str, dict]], dict[str, str]]:
    yearFilters = YearFilters.getForYear(year, **kwargs)
    modifiedMatchupIdToOwnerIdMap: dict = dict()
    teamForNames: dict[str, str] = dict()
    ownerForIds: dict[str, str] = dict()
    teamAgainstNames: dict[str, str] = dict()
    ownerAgainstIds: dict[str, str] = dict()
    teamForScores: dict[str, float | int] = dict()
    teamAgainstScores: dict[str, float | int] = dict()
    matchupTypes: dict[str, str] = dict()
    yearNumbers: dict[str, int] = dict()
    weekNumbers: dict[str, int] = dict()

    for week in year.weeks:
        if yearFilters.weekNumberStart <= week.weekNumber <= yearFilters.weekNumberEnd:
            for matchup in week.matchups:
                if matchup.matchupType in yearFilters.includeMatchupTypes \
                        and (matchup.multiWeekMatchupId is None or yearFilters.includeMultiWeekMatchups is True):
                    teamA = YearNavigator.getTeamById(year, matchup.teamAId)
                    teamB = YearNavigator.getTeamById(year, matchup.teamBId)
                    # add matchup for both teams
                    # add for "A" team
                    modifiedMatchupId = f"{matchup.id}A"
                    modifiedMatchupIdToOwnerIdMap[modifiedMatchupId] = teamA.ownerId
                    teamForNames[modifiedMatchupId] = teamA.name
                    ownerForIds[modifiedMatchupId] = teamA.ownerId
                    teamAgainstNames[modifiedMatchupId] = teamB.name
                    ownerAgainstIds[modifiedMatchupId] = teamB.ownerId
                    teamForScores[modifiedMatchupId] = matchup.teamAScore
                    teamAgainstScores[modifiedMatchupId] = matchup.teamBScore
                    matchupTypes[modifiedMatchupId] = matchup.matchupType.name
                    yearNumbers[modifiedMatchupId] = year.yearNumber
                    weekNumbers[modifiedMatchupId] = week.weekNumber
                    # add for "B" team
                    modifiedMatchupId = f"{matchup.id}B"
                    modifiedMatchupIdToOwnerIdMap[modifiedMatchupId] = teamB.ownerId
                    teamForNames[modifiedMatchupId] = teamB.name
                    ownerForIds[modifiedMatchupId] = teamB.ownerId
                    teamAgainstNames[modifiedMatchupId] = teamA.name
                    ownerAgainstIds[modifiedMatchupId] = teamA.ownerId
                    teamForScores[modifiedMatchupId] = matchup.teamBScore
                    teamAgainstScores[modifiedMatchupId] = matchup.teamAScore
                    matchupTypes[modifiedMatchupId] = matchup.matchupType.name
                    yearNumbers[modifiedMatchupId] = year.yearNumber
                    weekNumbers[modifiedMatchupId] = week.weekNumber

    titlesAndStatDicts = [
        ("Team For", teamForNames),
        ("Team Against", teamAgainstNames),
        ("Week Number", weekNumbers),
        ("Matchup Type", matchupTypes),
        ("Points For", teamForScores),
        ("Points Against", teamAgainstScores)
    ]
    if includeYears:
        titlesAndStatDicts.insert(2, ("Year", yearNumbers))
    if includeOwnerIds:
        titlesAndStatDicts.insert(1, ("Owner ID For", ownerForIds))
        titlesAndStatDicts.insert(3, ("Owner ID Against", ownerAgainstIds))

    return titlesAndStatDicts, modifiedMatchupIdToOwnerIdMap


def allTimeMatchupsStatSheet(league: League, **kwargs) -> tuple[list[tuple[str, dict]], dict[str, str]]:
    allTimeFilters = AllTimeFilters.getForLeague(league, **kwargs.copy())

    allYearMatchupStatSheets: list = list()
    allModifiedMatchupIdToOwnerIdMaps: list = list()
    teamForNames: dict[str, str] = dict()
    ownerForNames: dict[str, str] = dict()
    teamAgainstNames: dict[str, str] = dict()
    ownerAgainstNames: dict[str, str] = dict()
    teamForScores: dict[str, float | int] = dict()
    teamAgainstScores: dict[str, float | int] = dict()
    matchupTypes: dict[str, str] = dict()
    weekNumbers: dict[str, int] = dict()
    yearNumbers: dict[str, int] = dict()

    for year in league.years:
        if allTimeFilters.yearNumberStart <= year.yearNumber <= allTimeFilters.yearNumberEnd:
            currentYearMatchupStatSheet, currentModifiedMatchupIdToOwnerIdMap = yearMatchupsStatSheet(year,
                                                                                                      includeYears=True,
                                                                                                      includeOwnerIds=True,
                                                                                                      **kwargs.copy())
            allYearMatchupStatSheets.append(currentYearMatchupStatSheet)
            allModifiedMatchupIdToOwnerIdMaps.append(currentModifiedMatchupIdToOwnerIdMap)

    # combine responses into 1 response
    for yearMatchupStatSheet in allYearMatchupStatSheets:
        for title, statDict in yearMatchupStatSheet:
            if title == "Team For":
                teamForNames.update(statDict)
            elif title == "Team Against":
                teamAgainstNames.update(statDict)
            elif title == "Week Number":
                weekNumbers.update(statDict)
            elif title == "Matchup Type":
                matchupTypes.update(statDict)
            elif title == "Points For":
                teamForScores.update(statDict)
            elif title == "Points Against":
                teamAgainstScores.update(statDict)
            elif title == "Year":
                yearNumbers.update(statDict)
            elif title == "Owner ID For":
                # turn owner IDs into owner names
                currentOwnerForNames: dict[str, str] = dict()
                for key, ownerId in statDict.items():
                    owner = LeagueNavigator.getOwnerById(league, ownerId)
                    currentOwnerForNames[key] = owner.name
                ownerForNames.update(currentOwnerForNames)
            elif title == "Owner ID Against":
                # turn owner IDs into owner names
                currentOwnerAgainstNames: dict[str, str] = dict()
                for key, ownerId in statDict.items():
                    owner = LeagueNavigator.getOwnerById(league, ownerId)
                    currentOwnerAgainstNames[key] = owner.name
                ownerAgainstNames.update(currentOwnerAgainstNames)

    combinedModifiedMatchupIdToOwnerIdMap: dict = dict()
    for modifiedMatchupIdToOwnerIdMap in allModifiedMatchupIdToOwnerIdMaps:
        combinedModifiedMatchupIdToOwnerIdMap.update(modifiedMatchupIdToOwnerIdMap)

    return [("Team For", teamForNames),
            ("Owner For", ownerForNames),
            ("Team Against", teamAgainstNames),
            ("Owner Against", ownerAgainstNames),
            ("Year", yearNumbers),
            ("Week Number", weekNumbers),
            ("Matchup Type", matchupTypes),
            ("Points For", teamForScores),
            ("Points Against", teamAgainstScores)], combinedModifiedMatchupIdToOwnerIdMap



import espn_api.football as espn
from espn_api.football import League as ESPNLeague
from espn_api.football import Team as ESPNTeam

from leeger.enum.MatchupType import MatchupType
from leeger.league_loader.LeagueLoader import LeagueLoader
from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Owner import Owner
from leeger.model.league.Team import Team
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year
from leeger.validate import leagueValidation


class ESPNLeagueLoader(LeagueLoader):
    """
    Responsible for loading a League from ESPN Fantasy Football.
    https://www.espn.com/fantasy/football/
    """
    __ESPN_WIN_OUTCOME: str = "W"
    __ESPN_LOSS_OUTCOME: str = "L"
    __ESPN_TIE_OUTCOME: str = "T"
    __ESPN_BYE_OUTCOME: str = "U"
    __TEAMS_IN_PLAYOFFS_TO_PLAYOFF_WEEK_COUNT_MAP: dict[int, int] = {
        2: 1,
        3: 2,
        4: 2,
        6: 3,
        7: 3,
        8: 3
    }

    def __init__(self, leagueId: str, years: list[int], espnS2: str = None, swid: str = None, **kwargs):
        # validation
        try:
            int(leagueId)
        except ValueError:
            raise ValueError(f"League ID '{leagueId}' could not be turned into an int.")
        super().__init__(leagueId, years, **kwargs)

        self.__espnS2 = espnS2
        self.__swid = swid
        self.__espnTeamIdToTeamMap: dict[str, Team] = dict()
        self.__ownerNamesAndAliases: dict[str, list[str]] = dict()

    def loadLeague(self) -> League:
        espnLeagueYears = list()
        for year in self._years:
            espnLeagueYears.append(
                espn.League(league_id=int(self._leagueId), year=year, espn_s2=self.__espnS2, swid=self.__swid))

        league = self.__buildLeague(espnLeagueYears)
        # validate new league
        leagueValidation.runAllChecks(league)
        return league

    def __buildLeague(self, espnLeagues: list[ESPNLeague]) -> League:
        years = list()
        leagueName = None
        for espnLeague in espnLeagues:
            leagueName = espnLeague.settings.name if leagueName is None else leagueName
            self.__loadOwners(espnLeague.teams)
            years.append(self.__buildYear(espnLeague))
        return League(name=leagueName, owners=self._owners, years=years)

    def __loadOwners(self, espnTeams: list[ESPNTeam]) -> None:
        if self._owners is None:
            owners = list()
            for espnTeam in espnTeams:
                # get general owner name if there is one
                generalOwnerName = self._getGeneralOwnerNameFromGivenOwnerName(espnTeam.owner)
                ownerName = generalOwnerName if generalOwnerName is not None else espnTeam.owner
                owners.append(Owner(name=ownerName))
            self._owners = owners

    def __buildYear(self, espnLeague: ESPNLeague) -> Year:
        teams = self.__buildTeams(espnLeague.teams)
        weeks = self.__buildWeeks(espnLeague)
        return Year(yearNumber=espnLeague.year, teams=teams, weeks=weeks)

    def __buildWeeks(self, espnLeague: ESPNLeague) -> list[Week]:
        weeks = list()
        for i in range(espnLeague.current_week):  # current week seems to be the last week in the league
            # get each teams matchup for that week
            matchups = list()
            # to avoid adding matchups twice, we keep track of the ESPN team IDs that have already had a matchup added
            espnTeamIDsWithMatchups = list()
            for espnTeam in espnLeague.teams:
                # skip if we already have this team in a matchup
                # OR
                # this team is on a bye
                if espnTeam.team_id in espnTeamIDsWithMatchups or espnTeam.outcomes[i] == self.__ESPN_BYE_OUTCOME:
                    continue
                # team A is *this* team
                espnTeamA = espnTeam
                teamA = self.__espnTeamIdToTeamMap[espnTeam.team_id]
                teamAScore = espnTeam.scores[i]
                # team B is their opponent
                espnTeamB = espnTeam.schedule[i]
                teamB = self.__espnTeamIdToTeamMap[espnTeam.schedule[i].team_id]
                teamBScore = self.__getESPNTeamByESPNId(espnTeam.schedule[i].team_id, espnLeague.teams).scores[i]
                # figure out tiebreakers if there needs to be one
                teamAHasTiebreaker = teamAScore == teamBScore and espnTeamA.outcomes[i] == self.__ESPN_WIN_OUTCOME
                teamBHasTiebreaker = teamAScore == teamBScore and espnTeamB.outcomes[i] == self.__ESPN_WIN_OUTCOME
                matchupType = self.__getMatchupType(espnLeague, i + 1, espnTeamA.team_id)
                matchups.append(Matchup(teamAId=teamA.id,
                                        teamBId=teamB.id,
                                        teamAScore=teamAScore,
                                        teamBScore=teamBScore,
                                        teamAHasTiebreaker=teamAHasTiebreaker,
                                        teamBHasTiebreaker=teamBHasTiebreaker,
                                        matchupType=matchupType))
                espnTeamIDsWithMatchups.append(espnTeam.team_id)
                espnTeamIDsWithMatchups.append(espnTeam.schedule[i].team_id)
            if len(matchups) > 0:
                weeks.append(Week(weekNumber=i + 1, matchups=matchups))
        return weeks

    def __getMatchupType(self, espnLeague: ESPNLeague, weekNumber: int, espnTeamId: int) -> MatchupType:
        isPlayoffWeek = weekNumber > espnLeague.settings.reg_season_count
        if isPlayoffWeek:
            # figure out if this team made the playoffs
            playoffTeamCount = espnLeague.settings.playoff_team_count
            espnTeam = self.__getESPNTeamByESPNId(espnTeamId, espnLeague.teams)
            if playoffTeamCount >= espnTeam.standing:
                # this team made the playoffs
                # figure out if this is the last week of playoffs (the championship week)
                numberOfPlayoffWeeks = self.__TEAMS_IN_PLAYOFFS_TO_PLAYOFF_WEEK_COUNT_MAP[playoffTeamCount]
                # TODO: Raise specific exception here if not found in map
                if weekNumber == espnLeague.settings.reg_season_count + numberOfPlayoffWeeks:
                    # this is the championship week
                    # figure out if this team has lost in the playoffs before this week
                    playoffOutcomes = espnTeam.outcomes[-numberOfPlayoffWeeks:-1]
                    hasLostInPlayoffs = self.__ESPN_LOSS_OUTCOME in playoffOutcomes
                    if hasLostInPlayoffs:
                        return MatchupType.PLAYOFF
                    else:
                        return MatchupType.CHAMPIONSHIP
                else:
                    # this is a non-championship playoff week
                    return MatchupType.PLAYOFF
            else:
                # this matchup was played by teams that missed the playoffs during a playoff week.
                # ignore it.
                return MatchupType.IGNORE
        else:
            return MatchupType.REGULAR_SEASON

    @staticmethod
    def __getESPNTeamByESPNId(espnTeamId: int, espnTeams: list[ESPNTeam]) -> ESPNTeam:
        for espnTeam in espnTeams:
            if espnTeam.team_id == espnTeamId:
                return espnTeam

    def __buildTeams(self, espnTeams: list[ESPNTeam]) -> list[Team]:
        teams = list()
        for espnTeam in espnTeams:
            owner = self._getOwnerByName(espnTeam.owner)
            team = Team(ownerId=owner.id, name=espnTeam.team_name)
            teams.append(team)
            self.__espnTeamIdToTeamMap[espnTeam.team_id] = team
        return teams
